/*
 *   display.c
 *
 *   This file is part of Emu48
 *
 *   Copyright (C) 1995 Sebastien Carlier
 *
 */
#include "pch.h"
#ifdef	UNDER_CE
#include "resce.h"
#else
#include "resource.h"
#endif
#include "Emu48.h"
#include "kml.h"

#define LCD1_ROW    144
#define LCD2_ROW    288
#define LCD3_ROW    576

UINT    nBackgroundX = 0;
UINT    nBackgroundY = 0;
UINT    nBackgroundW = 0;
UINT    nBackgroundH = 0;
UINT    nLcdX = 0;
UINT    nLcdY = 0;
UINT    nLcdDoubled = 1;
UINT	nLcdDiv=1;
BOOL	bVertical=FALSE;
LPBYTE	pbyLcd;
HDC     hLcdDC = NULL;
HDC     hMainDC = NULL;
static HBITMAP hLcdBitmap;
static HBITMAP hMainBitmap;
static HBITMAP hOldLcdBitmap;
static HBITMAP hOldMainBitmap;

#define	MULDIV(x,y,z)	(((x)*(y))/(z))

#define B 0x00FFFFFF
#define W 0x00000000
#define I 0xFFFFFFFF
static struct
{
	BITMAPINFOHEADER Lcd_bmih;
	DWORD dwColor[64];
} bmiLcd =
{
	{0x28,0/*x*/,0/*y*/,1,8,BI_RGB,0,0,0,64,0},
	{
		W,B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,
		B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,
		I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,
		I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I
	}
};
#undef B
#undef W
#undef I

static DWORD Pattern[16];

VOID UpdateContrast(BYTE byContrast)
{
	DWORD c = byContrast;
	DWORD b = byContrast + 0x20;
	WORD i,j;
	if (bmiLcd.dwColor[b] == 0xFFFFFFFF) b = 0;

//	_ASSERT(nLcdDoubled == 1 || nLcdDoubled == 2 || nLcdDoubled == 4);

	for (i=0; i<16; ++i)
	{
		Pattern[i] = 0;
		for (j=8; j>0; j>>=1)
		{
			Pattern[i] = (Pattern[i] << 8) | ((i&j) ? c : b);
		}
	}
	return;
}

VOID SetLcdColor(UINT nId, UINT nRed, UINT nGreen, UINT nBlue)
{
	bmiLcd.dwColor[nId&0x3F] = ((nRed&0xFF)<<16)|((nGreen&0xFF)<<8)|(nBlue&0xFF);
	return;
}

VOID CreateLcdBitmap()
{
	// create LCD bitmap
//	_ASSERT(nLcdDoubled == 1 || nLcdDoubled == 2 || nLcdDoubled == 4);
	if (bVertical)
	{
		bmiLcd.Lcd_bmih.biHeight = LCD1_ROW;
		bmiLcd.Lcd_bmih.biWidth = 64;
	}
	else
	{
		bmiLcd.Lcd_bmih.biWidth = LCD1_ROW;
		bmiLcd.Lcd_bmih.biHeight = -64;
	}
	hLcdDC = CreateCompatibleDC(hWindowDC);
	hLcdBitmap = CreateDIBSection(hLcdDC, (BITMAPINFO*)&bmiLcd,DIB_RGB_COLORS, (LPVOID*)&pbyLcd, NULL, 0);
	hOldLcdBitmap = SelectObject(hLcdDC, hLcdBitmap);
	UpdateContrast(Chipset.contrast);
}

VOID DestroyLcdBitmap()
{
	WORD i;
	// clear background colors
	for (i=32; i< 64; ++i) bmiLcd.dwColor[i] = 0xFFFFFFFF;

	if (hLcdDC != NULL)
	{
		// destroy LCD bitmap
		SelectObject(hLcdDC, hOldLcdBitmap);
		DeleteObject(hLcdBitmap);
		DeleteDC(hLcdDC);
		hLcdDC = NULL;
		hLcdBitmap = NULL;
		hOldLcdBitmap = NULL;
	}
	return;
}

BOOL CreateMainBitmap(LPTSTR szFilename)
{
	hMainDC = CreateCompatibleDC(hWindowDC);
	hMainBitmap = LoadBitmapFile(szFilename);
	if (hMainBitmap == NULL)
	{
		DeleteDC(hMainDC);
		return FALSE;
	}
	hOldMainBitmap = SelectObject(hMainDC, hMainBitmap);
	SelectPalette(hMainDC, hPalette, FALSE);
	return TRUE;
}

VOID DestroyMainBitmap()
{
	if (hMainDC != NULL)
	{
		// destroy Main bitmap
		SelectObject(hMainDC, hOldMainBitmap);
		DeleteObject(hMainBitmap);
		DeleteDC(hMainDC);
		hMainDC = NULL;
		hMainBitmap = NULL;
		hOldMainBitmap = NULL;
	}
	return;
}

//****************
//*
//* LCD functions
//*
//****************

VOID UpdateDisplayPointers()
{
	// calculate display width
	Chipset.width = (34 + Chipset.loffset + (Chipset.boffset / 4) * 2) & 0xFFFFFFFE;
	Chipset.end1 = Chipset.start1 + (Chipset.lcounter + 1) * Chipset.width;
	if (Chipset.end1 < Chipset.start1)
	{
		// calculate first address of main display
		Chipset.start12 = Chipset.end1 - Chipset.width;
		// calculate last address of main display
		Chipset.end1 = Chipset.start1 - Chipset.width;
	}
	else
	{
		Chipset.start12 = Chipset.start1;
	}
	Chipset.end2 = Chipset.start2 + (63 - Chipset.lcounter) * 34;
}

static BYTE Buf[36];
static BOOL bScreenIsClean = FALSE;

__inline	long	DrawPixelVertical(DWORD pat,long l)
{
	BYTE p1=(BYTE)(pat&0xff);
	BYTE p2=(BYTE)((pat>>8)&0xff);
	BYTE p3=(BYTE)((pat>>16)&0xff);
	BYTE p4=(BYTE)((pat>>24)&0xff);
	pbyLcd[l]=p1;
	l+=64;
	pbyLcd[l]=p2;
	l+=64;
	pbyLcd[l]=p3;
	l+=64;
	pbyLcd[l]=p4;
	l+=64;
	return l;
}

VOID UpdateMainDisplay()
{
	UINT x, y;
	INT nLines;
	DWORD d = Chipset.start1;
	BYTE *p = pbyLcd;

	if (!pbyLcd) return;
	if (!Chipset.dispon)
	{
		nLines = 64;
		if (!bScreenIsClean)
		{
			bScreenIsClean = TRUE;
			ZeroMemory(pbyLcd, LCD1_ROW * nLines);
		}
	}
	else
	{
		nLines = Chipset.lcounter + 1;
		bScreenIsClean = FALSE;
//		_ASSERT(nLcdDoubled == 1 || nLcdDoubled == 2 || nLcdDoubled == 4);
		for (y=0; y<=Chipset.lcounter; y++)
		{
			Npeek(Buf,d,36);
			if (bVertical)
			{
				long l=y;
				for (x=0;x<36;x++)
				{
					l=DrawPixelVertical(Pattern[Buf[x]],l);
				}
			}
			else
			{
				for (x=0; x<36; x++) *(((DWORD*)p)++)=Pattern[Buf[x]];
			}
			d+=Chipset.width;
		}
		EnterCriticalSection(&csGDILock);		// solving NT GDI problems
		{
			if (bVertical)
				StretchBlt(hWindowDC, 
					nLcdX, nLcdY, 
					MULDIV(nLines,nLcdDoubled,nLcdDiv), MULDIV(131,nLcdDoubled,nLcdDiv), 
					hLcdDC, 
					0, Chipset.boffset+144-131, 
					nLines,131, SRCCOPY);
			else
				StretchBlt(hWindowDC, 
					nLcdX, nLcdY, 
					MULDIV(131,nLcdDoubled,nLcdDiv), 
					MULDIV(nLines,nLcdDoubled,nLcdDiv),
					hLcdDC, Chipset.boffset, 0, 131, nLines,SRCCOPY);
			GdiFlush();
		}
		LeaveCriticalSection(&csGDILock);
	}
	return;
}

VOID UpdateMenuDisplay()
{
	UINT x, y;
	BYTE *p;
	DWORD d = Chipset.start2;

	if (!pbyLcd) return;
	if (!Chipset.dispon) return;
	if (Chipset.lcounter==0x3F) return;		// menu disabled
//	_ASSERT(nLcdDoubled == 1 || nLcdDoubled == 2 || nLcdDoubled == 4);
	// calculate offset once
	p = pbyLcd + ((Chipset.lcounter+1)*LCD1_ROW);
	for (y=Chipset.lcounter+1; y<64; y++)
	{
		Npeek(Buf,d,34);				// 34 nibbles are viewed
		if (bVertical)
		{
			long l=y;
			for (x=0;x<36;x++)
			{
				l=DrawPixelVertical(Pattern[Buf[x]],l);
			}
		}
		else
		{
			for (x=0; x<36; x++) 
				*(((DWORD*)p)++)=Pattern[Buf[x]];
		}
		d+=34;
	}
	EnterCriticalSection(&csGDILock);		// solving NT GDI problems
	{
		if (bVertical)
		{
			StretchBlt(hWindowDC, nLcdX+MULDIV(Chipset.lcounter+1,nLcdDoubled,nLcdDiv), 
				nLcdY,
				MULDIV(63-Chipset.lcounter,nLcdDoubled,nLcdDiv),MULDIV(131,nLcdDoubled,nLcdDiv), 
				hLcdDC, 
				(Chipset.lcounter+1), Chipset.boffset+144-131, 
				(63-Chipset.lcounter),131, SRCCOPY);
		}
		else
		{
			StretchBlt(hWindowDC, nLcdX, 
				nLcdY+MULDIV(Chipset.lcounter+1,nLcdDoubled,nLcdDiv),
				MULDIV(131,nLcdDoubled,nLcdDiv), MULDIV(63-Chipset.lcounter,nLcdDoubled,nLcdDiv),
				hLcdDC, 
				0, (Chipset.lcounter+1), 131, (63-Chipset.lcounter),SRCCOPY);
		}
		GdiFlush();
	}
	LeaveCriticalSection(&csGDILock);
	return;
}

VOID WriteToMainDisplay(LPBYTE a, DWORD d, UINT s)
{
	INT	x0, x;
	INT	y0, y;
	DWORD	*p;

	INT lWidth = abs(Chipset.width);		// display width

	d -= Chipset.start1;					// nibble offset to DISPADDR (start of display)
	d += 64 * lWidth;						// make positive offset
	y0 = abs((INT) d / lWidth - 64);		// bitmap row
	x0 = (INT) d % lWidth;					// bitmap coloumn
	y = y0; x = x0;							// load loop variables

	// outside main display area
	_ASSERT(y0 >= 0 && y0 <= (INT) Chipset.lcounter);

	// illegal zoom factor
//	_ASSERT(nLcdDoubled == 1 || nLcdDoubled == 2 || nLcdDoubled == 4);

	// calculate memory position in LCD bitmap
	if (bVertical)
	{
		long l;

		l=x0*64*4+y0;
		while (s--)
		{
			if (x<36)
				DrawPixelVertical(Pattern[*a],l);
			++a;
			++x;
			if ((x==lWidth)&&s)					// end of display line
			{
				// end of main display area
				if (y == (INT) Chipset.lcounter) break;

				x = 0;							// first coloumn
				++y;							// next row
				l=y;
			}
			else 
			{
				l+=64*4;
			}
		}
		x0 = 0;								// no x-position offset
		x  = 131;							// redraw complete lines

		++y;								// redraw this line as well
	}
	else
	{
		p = (DWORD*) (pbyLcd + y0*LCD1_ROW
				     + x0*sizeof(bmiLcd.dwColor[0]));
		while (s--)								// loop for nibbles to write
		{
			if (x<36)							// only fill visible area
			{
				*p = Pattern[*a];
			}
			++a;								// next value to write
			++x;								// next x position
			if ((x==lWidth)&&s)					// end of display line
			{
				// end of main display area
				if (y == (INT) Chipset.lcounter) break;

				x = 0;							// first coloumn
				++y;							// next row
				// recalculate bitmap memory position of new line
				p = (DWORD*) (pbyLcd+y*LCD1_ROW);
			}
			else 
			{
				p ++;				// next x position in bitmap
			}
		}

		// update window region
		if (y0 != y)							// changed more than one line
		{
			x0 = 0;								// no x-position offset
			x  = 131;							// redraw complete lines

			++y;								// redraw this line as well
		}
		else
		{
			x0 <<= 2; x <<= 2;					// x-position in pixel
			_ASSERT(x >= x0);					// can't draw negative number of pixel 
			x -= x0;							// number of pixels to update

			x0 -= Chipset.boffset;				// adjust x-position with left margin
			if (x0 < 0) x0 = 0;
					
			if (x0   > 131) x0 = 131;			// cut right borders
			if (x+x0 > 131) x  = 131 - x0;

			y = y0 + 1;							// draw one line
		}

	}
	EnterCriticalSection(&csGDILock);
	{
		if (bVertical)
		{
			StretchBlt(hWindowDC, 
				nLcdX+MULDIV(y0,nLcdDoubled,nLcdDiv), nLcdY+MULDIV(x0,nLcdDoubled,nLcdDiv), 
				MULDIV(y-y0,nLcdDoubled,nLcdDiv), MULDIV(x,nLcdDoubled,nLcdDiv), 
				hLcdDC, 
				y0, x0+Chipset.boffset+144-131, 
				y-y0, x,
				SRCCOPY);
		}
		else
		{
			StretchBlt(hWindowDC, 
				nLcdX+MULDIV(x0,nLcdDoubled,nLcdDiv), nLcdY+MULDIV(y0,nLcdDoubled,nLcdDiv), 
				MULDIV(x,nLcdDoubled,nLcdDiv), MULDIV(y-y0,nLcdDoubled,nLcdDiv), 
				hLcdDC, 
				x0+Chipset.boffset, y0, 
				x, y-y0, 
				SRCCOPY);
		}
		GdiFlush();
	}
	LeaveCriticalSection(&csGDILock);
	return;
}

VOID WriteToMenuDisplay(LPBYTE a, DWORD d, UINT s)
{
	UINT x0, x;
	UINT y0, y;
	DWORD *p;

	UINT lWidth = abs(Chipset.width);		// display width

	if (Chipset.lcounter==0x3F) return;		// menu disabled
	d -= Chipset.start2;
	y0 = y = (d / 34) + (Chipset.lcounter+1);
	x0 = x = d % 34;
	if (x0 > 32) return;					// position out of viewed area



	// outside main display area
	_ASSERT(y0 >=  (INT) Chipset.lcounter);

	// illegal zoom factor
//	_ASSERT(nLcdDoubled == 1 || nLcdDoubled == 2 || nLcdDoubled == 4);

	// calculate memory position in LCD bitmap
	if (bVertical)
	{
		long l;

		l=x0*64*4+y0;
		while (s--)
		{
			if (x<36)
				DrawPixelVertical(Pattern[*a],l);
			++a;
			++x;
			if ((x==lWidth)&&s)					// end of display line
			{
				// end of main display area
				if (y == (INT) Chipset.lcounter) break;

				x = 0;							// first coloumn
				++y;							// next row
				l=y;
			}
			else 
			{
				l+=64*4;
			}
		}
		x0 = 0;								// no x-position offset
		x  = 131;							// redraw complete lines

		++y;								// redraw this line as well
	}
	else
	{
		p = (DWORD*) (pbyLcd + y0*LCD1_ROW
				     + x0*sizeof(bmiLcd.dwColor[0]));
		while (s--)								// loop for nibbles to write
		{
			if (x<36)							// only fill visible area
			{
				*p = Pattern[*a];
			}
			++a;								// next value to write
			++x;								// next x position
			if ((x==lWidth)&&s)					// end of display line
			{
				// end of main display area
				if (y == (INT) Chipset.lcounter) break;

				x = 0;							// first coloumn
				++y;							// next row
				// recalculate bitmap memory position of new line
				p = (DWORD*) (pbyLcd+y*LCD1_ROW);
			}
			else 
			{
				p ++;				// next x position in bitmap
			}
		}

		// update window region
		if (y0 != y)							// changed more than one line
		{
			x0 = 0;								// no x-position offset
			x  = 131;							// redraw complete lines

			++y;								// redraw this line as well
		}
		else
		{
			x0 <<= 2; x <<= 2;					// x-position in pixel
			_ASSERT(x >= x0);					// can't draw negative number of pixel 
			x -= x0;							// number of pixels to update

			x0 -= Chipset.boffset;				// adjust x-position with left margin
			if (x0 < 0) x0 = 0;
					
			if (x0   > 131) x0 = 131;			// cut right borders
			if (x+x0 > 131) x  = 131 - x0;

			y = y0 + 1;							// draw one line
		}

	}
	EnterCriticalSection(&csGDILock);
	{
		if (bVertical)
		{
			StretchBlt(hWindowDC, 
				nLcdX+MULDIV(y0,nLcdDoubled,nLcdDiv), nLcdY+MULDIV(x0,nLcdDoubled,nLcdDiv), 
				MULDIV(y-y0,nLcdDoubled,nLcdDiv), MULDIV(x,nLcdDoubled,nLcdDiv), 
				hLcdDC, 
				y0, x0+Chipset.boffset+144-131, 
				y-y0, x,
				SRCCOPY);
		}
		else
		{
			StretchBlt(hWindowDC, 
				nLcdX+MULDIV(x0,nLcdDoubled,nLcdDiv), nLcdY+MULDIV(y0,nLcdDoubled,nLcdDiv), 
				MULDIV(x,nLcdDoubled,nLcdDiv), MULDIV(y-y0,nLcdDoubled,nLcdDiv), 
				hLcdDC, 
				x0+Chipset.boffset, y0, 
				x, y-y0, 
				SRCCOPY);
		}
		GdiFlush();
	}
	LeaveCriticalSection(&csGDILock);
	return;
//	_ASSERT(nLcdDoubled == 1 || nLcdDoubled == 2 || nLcdDoubled == 4);
/*	if (bVertical)
	{
		long	l;
		l=x*LCD1_ROW+y;
		while (s--)
		{
			if (x<34)
				DrawPixelVertical(Pattern[*a],l);
			a++;
			x++;
			if ((x==34)&&s)
			{
				x=0;
				y++;
				if (y==64) break;
				l=y;
			} 
			else 
				l+=64*4;
		}
	}
	else
	{
//Esto es una cosa rara.
		p  = (DWORD*)(pbyLcd + y0*LCD1_ROW + x0*4);
		while (s--)
		{
			if (x<34) *p = Pattern[*a];
			a++;
			x++;
			if ((x==34)&&s)
			{
				x=0;
				y++;
				if (y==64) break;
				p=(DWORD*)(pbyLcd+y*LCD1_ROW);
			} else p++;
		}
	}
	EnterCriticalSection(&csGDILock);	// solving NT GDI problems
	{
		if (bVertical)
		{
			if (y0!=y)
			{
				StretchBlt(hWindowDC, 
					nLcdX+MULDIV(y0,nLcdDoubled,nLcdDiv), nLcdY, 
					MULDIV(y-y0+1,nLcdDoubled,nLcdDiv), MULDIV(131,nLcdDoubled,nLcdDiv), 
					hLcdDC, 
					y0, Chipset.boffset+144-131, 
					y-y0+1, 131,
					SRCCOPY);
			}
			else
			{
				x0<<=2; x<<=2;
				if (x>131) x=131;
				StretchBlt(hWindowDC, 
					nLcdX+MULDIV(y0,nLcdDoubled,nLcdDiv), nLcdY+MULDIV(x0,nLcdDoubled,nLcdDiv), 
					MULDIV(y-y0+1,nLcdDoubled,nLcdDiv), MULDIV(x-x0,nLcdDoubled,nLcdDiv), 
					hLcdDC, 
					y0, x0+Chipset.boffset+144-131, 
					y-y0+1, x-x0, 
					SRCCOPY);
			}
		}
		else
		{
			if (y0!=y)
			{
				StretchBlt(hWindowDC, 
					nLcdX, nLcdY+MULDIV(y0,nLcdDoubled,nLcdDiv), 
					MULDIV(131,nLcdDoubled,nLcdDiv), MULDIV(y-y0+1,nLcdDoubled,nLcdDiv), 
					hLcdDC, 
					0, y0, 
					131, y-y0+1, 
					SRCCOPY);
			}
			else
			{
				x0<<=2; x<<=2;
				if (x>131) x=131;
				StretchBlt(hWindowDC, 
					nLcdX+MULDIV(x0,nLcdDoubled,nLcdDiv), nLcdY+MULDIV(y0,nLcdDoubled,nLcdDiv), 
					MULDIV(x-x0,nLcdDoubled,nLcdDiv), MULDIV(y-y0+1,nLcdDoubled,nLcdDiv), 
					hLcdDC, 
					x0, y0, 
					x-x0, y-y0+1, 
					SRCCOPY);
			}
		}
		GdiFlush();
	}
	LeaveCriticalSection(&csGDILock);*/
	return;
}

VOID UpdateAnnunciators()
{
	BYTE c;
	
	c = (BYTE)(Chipset.IORam[0xB] | (Chipset.IORam[0xC]<<4));
	if (!(c&0x80)) c=0;
	DrawAnnunciator(1,c&0x01);
	DrawAnnunciator(2,c&0x02);
	DrawAnnunciator(3,c&0x04);
	DrawAnnunciator(4,c&0x08);
	DrawAnnunciator(5,c&0x10);
	DrawAnnunciator(6,c&0x20);
	return;
}

VOID ResizeWindow()
{
	RECT rectWindow;
	RECT rectClient;

	if (hWnd == NULL) return;				// return if window closed

	rectWindow.left   = 0;
	rectWindow.top    = 0;
	rectWindow.right  = nBackgroundW;
	rectWindow.bottom = nBackgroundH;
#ifndef	UNDER_CE
	AdjustWindowRect(&rectWindow, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);
#endif
#ifndef	UNDER_CE_PPC
	SetWindowPos (hWnd, (HWND)NULL, 0, 0,
		rectWindow.right  - rectWindow.left,
		rectWindow.bottom - rectWindow.top,
		SWP_NOMOVE | SWP_NOZORDER);
#endif
	GetClientRect(hWnd, &rectClient);
#ifndef	UNDER_CE
	AdjustWindowRect(&rectClient, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);
#endif
	if (rectClient.bottom < rectWindow.bottom)
	{
		rectWindow.bottom += (rectWindow.bottom - rectClient.bottom);
#ifndef	UNDER_CE_PPC
		SetWindowPos (hWnd, (HWND)NULL, 0, 0,
			rectWindow.right  - rectWindow.left,
			rectWindow.bottom - rectWindow.top,
			SWP_NOMOVE | SWP_NOZORDER);
#endif
	}
	InvalidateRect(hWnd,NULL,TRUE);
	return;
}