/*
 *   Emu48.c
 *
 *   This file is part of Emu48
 *
 *   Copyright (C) 1995 Sebastien Carlier
 *
 */
#include "pch.h"
#ifdef	UNDER_CE
#include "resce.h"
#include "wince.h"
#else
#include "resource.h"
#endif
#include "Emu48.h"
#include "io.h"
#include "kml.h"
#include "debugger.h"
#include "files.h"

#define MAXPORTS  16						// number of COM ports

#if defined(WIN32_PLATFORM_PSPC) && (_WIN32_WCE >= 300)
	#include <aygshell.h>
#endif

// #define MONOCHROME						// CF_BITMAP clipboard format

#ifdef _DEBUG
LPTSTR szNoTitle = _T("Emu48 1.23 Debug");
#else
LPTSTR szNoTitle = _T("Emu48 1.23");
#endif
LPTSTR szAppName = _T("Emu48");					// application name for DDE server
LPTSTR szTopic   = _T("Stack");					// topic for DDE server
LPTSTR szTitle   = NULL;
HMENU	hMenuCE=0;
HWND	hwndOwner=NULL;

static TCHAR szLicence[] =
	_T("This program is free software; you can redistribute it and/or modify\r\n\
it under the terms of the GNU General Public License as published by\r\n\
the Free Software Foundation; either version 2 of the License, or\r\n\
(at your option) any later version.\r\n\
This program is distributed in the hope that it will be useful,\r\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\r\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\r\n\
See the GNU General Public License for more details.\r\n\
You should have received a copy of the GNU General Public License\r\n\
along with this program; if not, write to the Free Software\r\n\
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.");


CRITICAL_SECTION csGDILock;					// critical section for hWindowDC
CRITICAL_SECTION csKeyLock;					// critical section for key scan
CRITICAL_SECTION csIOLock;					// critical section for I/O access
CRITICAL_SECTION csT1Lock;					// critical section for timer1 access
CRITICAL_SECTION csT2Lock;					// critical section for timer2 access
CRITICAL_SECTION csRecvLock;
LARGE_INTEGER    lFreq;						// counter frequency
LARGE_INTEGER	 lAppStart;
DWORD            idDdeInst;					// DDE server id
UINT             uCF_HpObj;					// DDE clipboard format
HINSTANCE        hApp = NULL;
HWND             hWnd = NULL;
HWND             hDlgDebug = NULL;			// handle for debugger dialog
HWND             hDlgFind = NULL;			// handle for debugger find dialog
HWND		 hwndCB = NULL;
HDC              hWindowDC  = NULL;
HPALETTE         hPalette = NULL;
HPALETTE         hOldPalette = NULL;		// old palette of hWindowDC
HANDLE           hThread;
DWORD            lThreadId;
HANDLE           hEventShutdn;				// event handle to stop cpu thread
BOOL             bAutoSave = FALSE;
BOOL             bAutoSaveOnExit = TRUE;
BOOL             bAlwaysDisplayLog = TRUE;
BOOL		 bDelayKey=1;

static BOOL      bRealSpeed = FALSE;

#ifdef	UNDER_CE
static TBBUTTON tbButtons[] = {
	{STD_FILENEW,    ID_FILE_NEW, TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0, 0, -1},
	{STD_FILEOPEN, ID_FILE_OPEN,	TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0, 0, -1},
	{STD_FILESAVE, ID_FILE_SAVE,	TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0, 0, -1},
	{0,              0,        TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0, 0, -1},
	{STD_PROPERTIES, ID_VIEW_SETTINGS,TBSTATE_ENABLED,  TBSTYLE_BUTTON, 0, 0, 0, -1},
	{STD_UNDO,ID_VIEW_SCRIPT,TBSTATE_ENABLED,  TBSTYLE_BUTTON, 0, 0, 0, -1},
	{STD_REPLACE,IDC_SWAP,		TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0, 0, -1},
	{STD_DELETE,ID_VIEW_RESET,		TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0, 0, -1},
};
static	int	nButtons=sizeof(tbButtons)/sizeof(TBBUTTON);
#endif

//################
//#
//#    Window Status
//#
//################


VOID SetWindowTitle(LPTSTR szString)
{
	if (szTitle)
	{
		LocalFree(szTitle);
	}
	if (szString)
	{
		szTitle = DuplicateString(szString);
	}
	else
	{
		szTitle = NULL;
	}
	_ASSERT(hWnd != NULL);
#ifndef	UNDER_CE_PPC
	if (szTitle)
	{
		SetWindowText(hWnd, szTitle);
	}
	else
	{
		SetWindowText(hWnd, szNoTitle);
	}
#endif
	return;
}

VOID UpdateWindowStatus()
{
#ifndef	UNDER_CE
	HMENU hMenu;

	if (hWnd == NULL) return;				// return if window closed

	if ((nState == 0)||(nState == 3))
	{
		if (szCurrentFilename[0])
			EnableMenuItem(hMenu,ID_FILE_SAVE,MF_ENABLED);
		else
			EnableMenuItem(hMenu,ID_FILE_SAVE,MF_GRAYED);
		EnableMenuItem(hMenu,ID_FILE_SAVEAS,MF_ENABLED);
		EnableMenuItem(hMenu,ID_FILE_CLOSE,MF_ENABLED);
		EnableMenuItem(hMenu,ID_BACKUP_SAVE,MF_GRAYED);
		EnableMenuItem(hMenu,ID_BACKUP_SAVE,MF_ENABLED);
		EnableMenuItem(hMenu,ID_VIEW_COPY,MF_ENABLED);
		EnableMenuItem(hMenu,ID_VIEW_RESET,MF_ENABLED);
		EnableMenuItem(hMenu,ID_OBJECT_LOAD,MF_ENABLED);
		EnableMenuItem(hMenu,ID_OBJECT_SAVE,MF_ENABLED);
		EnableMenuItem(hMenu,ID_STACK_COPY,MF_ENABLED);
		EnableMenuItem(hMenu,ID_STACK_PASTE,MF_ENABLED);
		EnableMenuItem(hMenu,ID_TOOL_DISASM,MF_ENABLED);
	}
	else
	{
		EnableMenuItem(hMenu,ID_FILE_SAVE,MF_GRAYED);
		EnableMenuItem(hMenu,ID_FILE_SAVEAS,MF_GRAYED);
		EnableMenuItem(hMenu,ID_FILE_CLOSE,MF_ENABLED);
		EnableMenuItem(hMenu,ID_BACKUP_SAVE,MF_GRAYED);
		EnableMenuItem(hMenu,ID_VIEW_COPY,MF_GRAYED);
		EnableMenuItem(hMenu,ID_VIEW_RESET,MF_GRAYED);
		EnableMenuItem(hMenu,ID_OBJECT_LOAD,MF_GRAYED);
		EnableMenuItem(hMenu,ID_OBJECT_SAVE,MF_GRAYED);
		EnableMenuItem(hMenu,ID_STACK_COPY,MF_GRAYED);
		EnableMenuItem(hMenu,ID_STACK_PASTE,MF_GRAYED);
		EnableMenuItem(hMenu,ID_TOOL_DISASM,MF_GRAYED);
	}
	if (bBackup)
	{
		EnableMenuItem(hMenu,ID_BACKUP_RESTORE,MF_ENABLED);
		EnableMenuItem(hMenu,ID_BACKUP_DELETE,MF_ENABLED);
	}
	else
	{
		EnableMenuItem(hMenu,ID_BACKUP_RESTORE,MF_GRAYED);
		EnableMenuItem(hMenu,ID_BACKUP_DELETE,MF_GRAYED);
	}
#endif
	return;
}



//################
//#
//#    Settings
//#
//################

// set listfield for serial combobox
static VOID SetCommList(HWND hDlg,int nIDDlgItem,LPTSTR szSetting)
{
	HANDLE hComm;
	BOOL   bAdd;
	WORD   wCount, wIndex = 1;
	TCHAR  szBuffer[16];

	WPARAM wSelect = 0;						// set select to disabled
	SendDlgItemMessage(hDlg,nIDDlgItem,CB_ADDSTRING,0,(LPARAM) NO_SERIAL);

	for (wCount = 1;wCount <= MAXPORTS;++wCount)
	{
		wsprintf(szBuffer,_T("COM%u"),wCount);
		if (lstrcmp(szBuffer,szSetting) == 0)
			wSelect = wIndex;
		// test if COM port is valid
		hComm = CreateFile(szBuffer,GENERIC_READ | GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
		if(hComm != INVALID_HANDLE_VALUE)
		{
			BOOL bError = CloseHandle(hComm);
			_ASSERT(bError);
			bAdd = TRUE;
		}

		if (bAdd)							// add item to combobox
		{
			SendDlgItemMessage(hDlg,nIDDlgItem,CB_ADDSTRING,0,(LPARAM) szBuffer);
			++wIndex;
		}
	}
	SendDlgItemMessage(hDlg,nIDDlgItem,CB_SETCURSEL,wSelect,0L);
}

static VOID ReadSettings()
{
//	InfoMessage(_T("In ReadSettings"));
	// Files
	GetPrivateProfileString(_T("Files"),_T("Emu48Directory"),szCurrentDirectory,szEmu48Directory,
		                    sizeof(szEmu48Directory)/sizeof(TCHAR),EMU48_INI);
//	InfoMessage(_T("Get Emu48Directory"));
	if (!szEmu48Directory[0])
	{
		TCHAR	szModuleName[260];
		int	i;

//		InfoMessage(_T("No Emu48Directory: Get ModuleFileName"));
		GetModuleFileName(hApp,szModuleName,260);
		for (i=_tcslen(szModuleName)-1;i>=0 && szModuleName[i]!=(TCHAR)'\\';i--)
			;
		szModuleName[i]=0;
		_tcscpy(szCurrentDirectory,szModuleName);
		_tcscpy(szEmu48Directory,szCurrentDirectory);
		WritePrivateProfileString(_T("Files"),_T("Emu48Directory"),szEmu48Directory,EMU48_INI);
	}
	bAutoSave       = GetPrivateProfileInt(_T("Files"),_T("AutoSave"),bAutoSave,EMU48_INI);
//	InfoMessage(_T("Got AutoSave"));
	bAutoSaveOnExit = GetPrivateProfileInt(_T("Files"),_T("AutoSaveOnExit"),bAutoSaveOnExit,EMU48_INI);
	// Port2
//	InfoMessage(_T("Got AutoSaveOnExit"));
	bPort2IsShared  = GetPrivateProfileInt(_T("Port2"),_T("IsShared"),0,EMU48_INI);
	GetPrivateProfileString(_T("Port2"),_T("Filename"),_T("SHARED.BIN"),szPort2Filename,sizeof(szPort2Filename),EMU48_INI);
	// KML
	bAlwaysDisplayLog = GetPrivateProfileInt(_T("KML"),_T("AlwaysDisplayLog"),bAlwaysDisplayLog,EMU48_INI);
//For Windows CE the key press can be delayed.
#ifdef	UNDER_CE
	bDelayKey=GetPrivateProfileInt(_T("Files"),_T("Delay"),1,EMU48_INI);
#endif
	// Emulator
	bRealSpeed = GetPrivateProfileInt(_T("Emulator"),_T("RealSpeed"),0,EMU48_INI);
	dwSXCycles = GetPrivateProfileInt(_T("Emulator"),_T("SXCycles"),82,EMU48_INI);
	dwGXCycles = GetPrivateProfileInt(_T("Emulator"),_T("GXCycles"),123,EMU48_INI);
//	InfoMessage(_T("Before SetSpeed"));
	SetSpeed(bRealSpeed);					// set speed
	// Serial
	GetPrivateProfileString(_T("Serial"),_T("Wire"),NO_SERIAL,szSerialWire,sizeof(szSerialWire),EMU48_INI);
//	InfoMessage(_T("Got Serial Wire"));
	GetPrivateProfileString(_T("Serial"),_T("Ir"),NO_SERIAL,szSerialIr,sizeof(szSerialIr),EMU48_INI);
//	InfoMessage(_T("Got Serial IR"));
	return;
}

static VOID WriteSettings()
{
	// Files
	WritePrivateProfileString(_T("Files"),_T("Emu48Directory"),szEmu48Directory,EMU48_INI);
	WritePrivateProfileInt(_T("Files"),_T("AutoSave"),bAutoSave,EMU48_INI);
	WritePrivateProfileInt(_T("Files"),_T("AutoSaveOnExit"),bAutoSaveOnExit,EMU48_INI);
	// Port2
	WritePrivateProfileInt(_T("Port2"),_T("IsShared"),bPort2IsShared,EMU48_INI);
	WritePrivateProfileString(_T("Port2"),_T("Filename"),szPort2Filename,EMU48_INI);
	// KML
	WritePrivateProfileInt(_T("KML"),_T("AlwaysDisplayLog"),bAlwaysDisplayLog,EMU48_INI);
	// Emulator
	WritePrivateProfileInt(_T("Emulator"),_T("RealSpeed"),bRealSpeed,EMU48_INI);
	WritePrivateProfileInt(_T("Emulator"),_T("SXCycles"),dwSXCycles,EMU48_INI);
	WritePrivateProfileInt(_T("Emulator"),_T("GXCycles"),dwGXCycles,EMU48_INI);
	// Serial
	WritePrivateProfileString(_T("Serial"),_T("Wire"),szSerialWire,EMU48_INI);
	WritePrivateProfileString(_T("Serial"),_T("Ir"),szSerialIr,EMU48_INI);
#ifdef	UNDER_CE
	WritePrivateProfileInt(_T("Files"),_T("Delay"),bDelayKey,EMU48_INI);
#endif
	return;
}

static BOOL CALLBACK SettingsProc(HWND hDlg, UINT message, DWORD wParam, LONG lParam)
{
	switch (message)
	{
	case WM_INITDIALOG:
		// init speed checkbox
		CheckDlgButton(hDlg,IDC_REALSPEED,bRealSpeed);
		CheckDlgButton(hDlg,IDC_AUTOSAVE,bAutoSave);
		CheckDlgButton(hDlg,IDC_AUTOSAVEONEXIT,bAutoSaveOnExit);
		CheckDlgButton(hDlg,IDC_ALWAYSDISPLOG,bAlwaysDisplayLog);

#ifndef	UNDER_CE
		// set disassebler mode
		CheckDlgButton(hDlg,(disassembler_mode == HP_MNEMONICS) ? IDC_DISASM_HP : IDC_DISASM_CLASS,BST_CHECKED);
#endif

		// set combobox parameter
		SetCommList(hDlg,IDC_WIRE,szSerialWire); 
		SetCommList(hDlg,IDC_IR,szSerialIr);
		if (CommConnect() != PORT_CLOSE)	// disable when port open
		{
			EnableWindow(GetDlgItem(hDlg,IDC_WIRE),FALSE);
			EnableWindow(GetDlgItem(hDlg,IDC_IR),FALSE);
		}

		// HP48SX/GX
		if (cCurrentRomType!='6' && cCurrentRomType!='A' && cCurrentRomType!='X')
		{
			// init port1 enable checkbox
			CheckDlgButton(hDlg,IDC_PORT1EN,(Chipset.cards_status & PORT1_PRESENT) != 0);
			CheckDlgButton(hDlg,IDC_PORT1WR,(Chipset.cards_status & PORT1_WRITE) != 0);
			// init port2 shared checkbox and set port2 filename
			CheckDlgButton(hDlg,IDC_PORT2ISSHARED,bPort2IsShared);
			SetDlgItemText(hDlg,IDC_PORT2,szPort2Filename);
			if (nState == SM_INVALID)				// Invalid State
			{
				EnableWindow(GetDlgItem(hDlg,IDC_PORT1EN),FALSE);
				EnableWindow(GetDlgItem(hDlg,IDC_PORT1WR),FALSE);
			}
		}
		else								// HP38G/HP49G
		{
			EnableWindow(GetDlgItem(hDlg,IDC_PORT1EN),FALSE);
			EnableWindow(GetDlgItem(hDlg,IDC_PORT1WR),FALSE);
			if (nState == 1)				// Invalid State
			{
			    // init port2 shared checkbox and set port2 filename
				CheckDlgButton(hDlg,IDC_PORT2ISSHARED,bPort2IsShared);
				SetDlgItemText(hDlg,IDC_PORT2,szPort2Filename);
			}
			if (cCurrentRomType=='X')		// HP49G
			{
				SendDlgItemMessage(hDlg,IDC_IR,CB_RESETCONTENT,0,0);
				EnableWindow(GetDlgItem(hDlg,IDC_IR),FALSE);
			}
		}
		if (nState != SM_INVALID)					// not in Invalid State
		{
			// hide port2 settings
			EnableWindow(GetDlgItem(hDlg,IDC_PORT2ISSHARED),FALSE);
			EnableWindow(GetDlgItem(hDlg,IDC_PORT2),FALSE);
		}
#ifdef	UNDER_CE
		CheckDlgButton(hDlg,IDC_KEYDELAY,bDelayKey);
		FULLSCREEN(hDlg);
#endif
		return TRUE;
	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK)
		{
			if (Chipset.Port1Size && cCurrentRomType!='X')
			{
				UINT nOldState = SwitchToState(3);
				// save old card status
				BYTE bCardsStatus = Chipset.cards_status;

				// port1 disabled?
				Chipset.cards_status &= ~(PORT1_PRESENT | PORT1_WRITE);
				if (IsDlgButtonChecked(hDlg, IDC_PORT1EN))
				{
					Chipset.cards_status |= PORT1_PRESENT;
					if (IsDlgButtonChecked(hDlg, IDC_PORT1WR))
						Chipset.cards_status |= PORT1_WRITE;
				}

				// changed card status in slot1?
				if (   ((bCardsStatus ^ Chipset.cards_status) & (PORT1_PRESENT | PORT1_WRITE)) != 0
					&& (Chipset.IORam[CARDCTL] & ECDT) != 0 && (Chipset.IORam[TIMER2_CTRL] & RUN) != 0
				   )
				{

					Chipset.HST |= MP;			// set Module Pulled
					IOBit(SRQ2,NINT,FALSE);		// set NINT to low
					Chipset.SoftInt = TRUE;		// set interrupt
					bInterrupt = TRUE;
				}
				Map(0x00,0xFF);
				SwitchToState(nOldState);
			}
			// get speed checkbox value
			bRealSpeed = IsDlgButtonChecked(hDlg,IDC_REALSPEED);
			bAutoSave = IsDlgButtonChecked(hDlg, IDC_AUTOSAVE);
			bAutoSaveOnExit = IsDlgButtonChecked(hDlg, IDC_AUTOSAVEONEXIT);
			bAlwaysDisplayLog = IsDlgButtonChecked(hDlg, IDC_ALWAYSDISPLOG);
			SetSpeed(bRealSpeed);			// set speed
			if (nState == SM_INVALID)
			{
				bPort2IsShared = IsDlgButtonChecked(hDlg,IDC_PORT2ISSHARED);
				GetDlgItemText(hDlg,IDC_PORT2,szPort2Filename,sizeof(szPort2Filename));
			}
#ifndef	UNDER_CE
			// set disassebler mode
			disassembler_mode = IsDlgButtonChecked(hDlg,IDC_DISASM_HP) ? HP_MNEMONICS : CLASS_MNEMONICS;
#endif
#ifdef	UNDER_CE
			bDelayKey=IsDlgButtonChecked(hDlg,IDC_KEYDELAY);
#endif
			// set combobox parameter
			GetDlgItemText(hDlg,IDC_WIRE,szSerialWire,sizeof(szSerialWire));
			if (cCurrentRomType!='X')		// HP49G Ir port is not connected
				GetDlgItemText(hDlg,IDC_IR,szSerialIr,sizeof(szSerialIr));
			EndDialog(hDlg, wParam);
		}
		if (LOWORD(wParam) == IDCANCEL)
		{
			EndDialog(hDlg, wParam);
		}
		break;
	}
	return FALSE;
    UNREFERENCED_PARAMETER(lParam);
}


//################
//#
//#    Save Helper
//#
//################

//
// UINT SaveChanges(BOOL bAuto);
// Return code :
// IDYES    File successfuly saved
// IDNO     File not saved
// IDCANCEL Cancel command
//
static UINT SaveChanges(BOOL bAuto)
{
	UINT uReply;
	
	if (pbyRom == NULL) return IDNO;
	
	if (bAuto)
		uReply = IDYES;
	else
		uReply = YesNoCancelMessage(_T("Do you want to save changes ?"));
	
	if (uReply != IDYES) return uReply;
	
	if (szCurrentFilename[0]==0)
	{ // Save As...
		uReply = GetSaveAsFilename();
		if (uReply != IDOK) return uReply;
		if (!SaveDocumentAs(szBufferFilename)) return IDCANCEL;
		WritePrivateProfileString(_T("Files"),_T("LastDocument"),szCurrentFilename,EMU48_INI);
		return IDYES;
	}
	
	SaveDocument();
	return IDYES;
}



//################
//#
//#    Message Handlers
//#
//################
void	CreateCommandBar()
{
#ifdef UNDER_CE
	BOOL bRes;

#if defined(WIN32_PLATFORM_PSPC) && (_WIN32_WCE >= 300)
        SHMENUBARINFO mbi;

	if (hwndCB)
		DestroyWindow(hwndCB);

    memset (&mbi, 0, sizeof (SHMENUBARINFO));
    mbi.cbSize     = sizeof (SHMENUBARINFO);
    mbi.hwndParent = hWnd;
    mbi.nToolBarId = IDR_MENU;
    mbi.hInstRes   = hApp;
    mbi.nBmpId     = 0;
    mbi.cBmpImages = 0;    

    if (!SHCreateMenuBar(&mbi))
        {
            MessageBox(hWnd, L"SHCreateMenuBar Failed", L"Error", MB_OK);
        }
    
    hwndCB = mbi.hwndMB;
#else
	// Set up the command bar...
	if (hwndCB)
		DestroyWindow(hwndCB);
	hwndCB = CommandBar_Create(hApp, hWnd, 1);
	// Insert the menu bar.
	bRes = CommandBar_InsertMenubarEx(hwndCB, hApp, MAKEINTRESOURCE(IDR_MENU), 0);
    CommandBar_AddBitmap(hwndCB, HINST_COMMCTRL, IDB_STD_SMALL_COLOR, 15, 16, 16);
	CommandBar_AddButtons(hwndCB,nButtons,tbButtons);
#endif
	// Add the Close button.
#ifndef	UNDER_CE_PPC
	bRes = CommandBar_AddAdornments(hwndCB, 0, 0);
#endif
	// Draw it...
	bRes = CommandBar_DrawMenuBar(hwndCB, 0);
#endif // UNDER_CE
}

//
// WM_CREATE
//
static LRESULT OnCreate(HWND hWindow)
{
   hWnd=hWindow;
   CreateCommandBar();
   if (GetPrivateProfileInt(_T("Files"),_T("Menu"),1,EMU48_INI))
   {
	hwndOwner = hWindow;
	hWnd = hWindow;
	hWindowDC = GetDC(hWnd);

#ifndef	UNDER_CE
	DragAcceptFiles(hWnd,TRUE);				// support dropped files
#endif
   }
   return 0;
}

//
// WM_DESTROY
//
static LRESULT OnDestroy(HWND hWindow)
{
#ifndef	UNDER_CE
	DragAcceptFiles(hWnd,FALSE);			// no WM_DROPFILES message any more
#endif
	SwitchToState(2);
	WriteSettings();
	ReleaseDC(hWnd, hWindowDC);
	SetWindowTitle(NULL);					// free memory of title
	hWindowDC = NULL;						// hWindowDC isn't valid any more
	hWnd = NULL;
	PostQuitMessage(0);
	return 0;
	UNREFERENCED_PARAMETER(hWindow);
}

//
// WM_PAINT
//
static LRESULT OnPaint(HWND hWindow)
{
	PAINTSTRUCT Paint;
	HDC hPaintDC;
	
	hPaintDC = BeginPaint(hWindow, &Paint);
	if (hMainDC != NULL)
	{
		BitBlt(hPaintDC, 0, 0, nBackgroundW, nBackgroundH, hMainDC, nBackgroundX, nBackgroundY, SRCCOPY);
		if ((nState==0)||(nState==3))
		{
			UpdateMainDisplay();
			UpdateMenuDisplay();
			UpdateAnnunciators();
			RefreshButtons();
		}
	}
	else
	{
		HPEN	greenPen,ap;
		HBRUSH	greenBrush,ab;
		RECT	r;
		GetClientRect(hWnd,&r);
		greenPen=CreatePen(PS_SOLID,1,RGB(0,0X80,0));
		greenBrush=CreateSolidBrush(RGB(0,0X80,0));
		ap=SelectObject(hPaintDC,greenPen);
		ab=SelectObject(hPaintDC,greenBrush);
		SetBkMode(hPaintDC,TRANSPARENT);
		Rectangle(hPaintDC,r.left,r.top,r.right-r.left,r.bottom-r.top);
		DrawText(hPaintDC,szLicence,-1,&r,DT_CENTER | DT_WORDBREAK);
		SelectObject(hPaintDC,ap);
		SelectObject(hPaintDC,ab);
		DeleteObject(greenPen);
		DeleteObject(greenBrush);
	}
	EndPaint(hWindow, &Paint);
	return 0;
}

//
// WM_DROPFILES
//
static LRESULT OnDropFiles(HANDLE hFilesInfo)
{
#ifndef	UNDER_CE
	char szFileName[MAX_PATH];
	WORD wNumFiles,wIndex;
	BOOL bSuccess;

	// get number of files dropped
	wNumFiles = DragQueryFile (hFilesInfo,(UINT)-1,NULL,0);

	if (!Chipset.dispon)					// calculator off, turn on
	{
		// turn on HP
		KeyboardEvent(TRUE,0,0x8000);
		Sleep(200);
		KeyboardEvent(FALSE,0,0x8000);
	}

	_ASSERT(nState == 0);					// Emulator must be in RUN state
	if (WaitForSleepState())				// wait for cpu SHUTDN then sleep state
	{
		DragFinish (hFilesInfo);
		InfoMessage("The emulator is busy.");
		return 0;
	}

	while (nState!=nNextState) Sleep(0);
	_ASSERT(nState==3);

	// get each name and load it into the emulator
	for (wIndex = 0;wIndex < wNumFiles;++wIndex)
	{
		DragQueryFile (hFilesInfo,wIndex,(LPSTR) szFileName,sizeof(szFileName));

		// szFileName has file name, now try loading it
		if ((bSuccess = LoadObject(szFileName)) == FALSE)
			break;
	}

	DragFinish (hFilesInfo);
	SwitchToState(0);						// run state
	while (nState!=nNextState) Sleep(0);
	_ASSERT(nState==0);

	if (bSuccess == FALSE)					// data not copied
		return 0;

	KeyboardEvent(TRUE,0,0x8000);
	Sleep(200);
	KeyboardEvent(FALSE,0,0x8000);
	// wait for sleep mode
	while(Chipset.Shutdn == FALSE) Sleep(0);
#endif
	return 0;
}

//
// ID_FILE_NEW
//
static LRESULT OnFileNew()
{
	UINT uReply;
	
	SaveBackup();
	if (pbyRom)
	{
		SwitchToState(1);
		uReply = SaveChanges(bAutoSave);
		if (uReply==IDCANCEL) goto cancel;
	}
	NewDocument();
	SetWindowTitle(_T("Untitled"));
	UpdateWindowStatus();
cancel:
	if (pbyRom) SwitchToState(0);
	return 0;
}

//
// ID_FILE_OPEN
//
LRESULT OnFileOpen()
{
	UINT uReply;
	
	if (pbyRom)
	{
		SwitchToState(1);
		uReply = SaveChanges(bAutoSave);
		if (uReply==IDCANCEL) goto cancel;
	}
	if (GetOpenFilename())
	{
		OpenDocument(szBufferFilename);
	}
cancel:
	if (pbyRom)
	{
		SwitchToState(0);
	}
	return 0;
}

//
// ID_FILE_SAVE
//
static LRESULT OnFileSave()
{
	if (pbyRom == NULL) return 0;
	SwitchToState(1);
	SaveChanges(TRUE);
	SwitchToState(0);
	return 0;
}

//
// ID_FILE_SAVEAS
//
static LRESULT OnFileSaveAs()
{
	UINT uReply;
	
	if (pbyRom == NULL) return 0;
	SwitchToState(1);
	
	uReply = GetSaveAsFilename();
	if (uReply != IDOK)
	{
		SwitchToState(0);
		return 0;
	}
	if (!SaveDocumentAs(szBufferFilename))
	{
		SwitchToState(0);
		return 0;
	}
	WritePrivateProfileString(_T("Files"),_T("LastDocument"),szCurrentFilename,EMU48_INI);
	
	SwitchToState(0);
	return 0;
}

//
// ID_FILE_CLOSE
//
static LRESULT OnFileClose()
{
	if (pbyRom == NULL) return 0;
	SwitchToState(1);
	if (SaveChanges(bAutoSave)!=IDCANCEL)
	{
		KillKML();
		ResetDocument();
		SetWindowTitle(NULL);
	}
	else
	{
		SwitchToState(0);
	}
	return 0;
}

//
// ID_FILE_EXIT
//
// WM_SYS_CLOSE
//
static LRESULT OnFileExit()
{
	SwitchToState(1);						// hold emulation thread
	if (SaveChanges(bAutoSaveOnExit) == IDCANCEL)
	{
		SwitchToState(0);					// on cancel restart emulation thread
		return 0;
	}
	DestroyWindow(hWnd);
	return 0;
}

//
// ID_STACK_COPY
//
static LRESULT OnStackCopy()				// copy data from stack
{
	HANDLE hClipObj;
	LPBYTE lpData;
	DWORD  dwAddress,dwSize;

	_ASSERT(nState == 0);					// Emulator must be in RUN state
	if (WaitForSleepState())				// wait for cpu SHUTDN then sleep state
	{
		InfoMessage(_T("The emulator is busy."));
		return 0;
	}

	while (nState!=nNextState) Sleep(0);
	_ASSERT(nState==3);

	if ((dwAddress = RPL_Pick(1)) == 0)		// pick address of level1 object
	{
		MessageBeep(MB_OK);					// error beep
		goto error;
	}

	if (Read5(dwAddress) != 0x02A2C)		// string object
	{
		MessageBeep(MB_OK);					// error beep
		goto error;
	}

	dwAddress += 5;							// address of string length
	dwSize = (Read5(dwAddress) - 5) / 2;	// length of string

	// memory allocation for clipboard data
	if ((hClipObj = LocalAlloc(LPTR,dwSize + 1)) == NULL)
		goto error;

	if (lpData = LocalLock(hClipObj))		// lock memory
	{
		// copy data into clipboard buffer
		for (dwAddress += 5;dwSize-- > 0;dwAddress += 2,++lpData)
			*lpData = Read2(dwAddress);
		*lpData = 0;						// set end of string

		LocalUnlock(hClipObj);				// unlock memory

		if (OpenClipboard(hWnd))
		{
			if (EmptyClipboard())
				SetClipboardData(CF_TEXT,hClipObj);
			else
				LocalFree(hClipObj);
			CloseClipboard();
		}
	}
	else									// lock memory failed
	{
		LocalFree(hClipObj);
	}

error:
	SwitchToState(0);
	return 0;
}

//
// ID_STACK_PASTE
//
static LRESULT OnStackPaste()				// paste data to stack
{
	HANDLE hClipObj;
	LPBYTE lpClipdata,lpData;

	BOOL   bSuccess = FALSE;

	if (!Chipset.dispon)					// calculator off, turn on
	{
		KeyboardEvent(TRUE,0,0x8000);
		Sleep(200);
		KeyboardEvent(FALSE,0,0x8000);
		// wait for sleep mode
		while(Chipset.Shutdn == FALSE) Sleep(0);
	}

	_ASSERT(nState == 0);					// Emulator must be in RUN state
	if (WaitForSleepState())				// wait for cpu SHUTDN then sleep state
	{
		InfoMessage(_T("The emulator is busy."));
		return 0;
	}

	while (nState!=nNextState) Sleep(0);
	_ASSERT(nState==3);

	if (OpenClipboard(hWnd))
	{
		if (   IsClipboardFormatAvailable(CF_TEXT)
			|| IsClipboardFormatAvailable(CF_OEMTEXT))
		{
			if (hClipObj = GetClipboardData(CF_TEXT))
			{
				if (lpClipdata = LocalLock(hClipObj))
				{
					DWORD dwSize = strlen(lpClipdata);
					if (lpData = (LPBYTE) LocalAlloc(LMEM_FIXED,dwSize * 2))
					{
						memcpy(lpData+dwSize,lpClipdata,dwSize); // copy data
						bSuccess = (WriteStack(lpData,dwSize) == S_ERR_NO);
						LocalFree(lpData);
					}
					LocalUnlock(hClipObj);
				}
			}
		}
		else								// unknown clipboard format
		{
			MessageBeep(MB_OK);				// error beep
		}
		CloseClipboard();
	}

	SwitchToState(0);						// run state
	while (nState!=nNextState) Sleep(0);
	_ASSERT(nState == 0);

	if (bSuccess == FALSE)					// data not copied
		return 0;

	KeyboardEvent(TRUE,0,0x8000);
	Sleep(200);
	KeyboardEvent(FALSE,0,0x8000);
	// wait for sleep mode
	while(Chipset.Shutdn == FALSE) Sleep(0);
	return 0;
}

//
// ID_VIEW_COPY
//
static LRESULT OnViewCopy()
{
	if (OpenClipboard(hWnd))
	{
		if (EmptyClipboard())
		{
//#if !defined MONOCHROME
#ifndef	UNDER_CE
			// DIB bitmap
			#define WIDTHBYTES(bits) (((bits) + 31) / 32 * 4)
			#define PALVERSION       0x300

			BITMAP bm;
			LPBITMAPINFOHEADER lpbi;
			PLOGPALETTE ppal;
			HBITMAP hBmp;
			HDC hBmpDC;
			HANDLE hClipObj;
			WORD wBits;
			DWORD dwLen, dwSizeImage;

			_ASSERT(nLcdDoubled == 1 || nLcdDoubled == 2 || nLcdDoubled == 4);
			hBmp = CreateCompatibleBitmap(hLcdDC,131*nLcdDoubled,64*nLcdDoubled);
			hBmpDC = CreateCompatibleDC(hLcdDC);
			hBmp = SelectObject(hBmpDC,hBmp);
			BitBlt(hBmpDC, 0, 0, 131*nLcdDoubled, 64*nLcdDoubled, hLcdDC, 0, 0, SRCCOPY);
			hBmp = SelectObject(hBmpDC,hBmp);

			// fill BITMAP structure for size information
			GetObject(hBmp, sizeof(bm), &bm);

			wBits = bm.bmPlanes * bm.bmBitsPixel;
			// make sure bits per pixel is valid 
			if (wBits <= 1)
				wBits = 1;
			else if (wBits <= 4)
				wBits = 4;
			else if (wBits <= 8)
				wBits = 8;
			else // if greater than 8-bit, force to 24-bit 
				wBits = 24;

			dwSizeImage = WIDTHBYTES((DWORD)bm.bmWidth * wBits) * bm.bmHeight;

			// calculate memory size to store CF_DIB data
			dwLen = sizeof(BITMAPINFOHEADER) + dwSizeImage;
			if (wBits != 24)				// a 24 bitcount DIB has no color table
			{
				// add size for color table
				dwLen += (1 << wBits) * sizeof(RGBQUAD);
			}

			// memory allocation for clipboard data
			if ((hClipObj = LocalAlloc(LPTR, dwLen)) != NULL)
			{
				lpbi = LocalLock(hClipObj);
				// initialize BITMAPINFOHEADER
				lpbi->biSize = sizeof(BITMAPINFOHEADER);
				lpbi->biWidth = bm.bmWidth;
				lpbi->biHeight = bm.bmHeight;
				lpbi->biPlanes = 1;
				lpbi->biBitCount = wBits;
				lpbi->biCompression = BI_RGB;
				lpbi->biSizeImage = dwSizeImage;
				lpbi->biXPelsPerMeter = 0;
				lpbi->biYPelsPerMeter = 0;
				lpbi->biClrUsed = 0;
				lpbi->biClrImportant = 0;
				// get bitmap color table and bitmap data
				GetDIBits(hBmpDC, hBmp, 0, lpbi->biHeight, (LPBYTE)lpbi + dwLen - dwSizeImage,
                          (LPBITMAPINFO)lpbi, DIB_RGB_COLORS);
				LocalUnlock(hClipObj);
				SetClipboardData(CF_DIB, hClipObj); 

				// get number of entries in the logical palette 
				GetObject(hPalette,sizeof(WORD),&wBits);

				// memory allocation for temporary palette data
				if ((ppal = LocalAlloc(LPTR, sizeof(LOGPALETTE) + wBits * sizeof(PALETTEENTRY))) != NULL)
				{
					ppal->palVersion    = PALVERSION;
					ppal->palNumEntries = wBits;
					GetPaletteEntries(hPalette, 0, wBits, ppal->palPalEntry);
					SetClipboardData(CF_PALETTE, CreatePalette(ppal)); 
					LocalFree(ppal);
				}
			}
			DeleteDC(hBmpDC);
			DeleteObject(hBmp);
			#undef WIDTHBYTES
			#undef PALVERSION
#else
			HBITMAP hOldBmp, hBmp;
			HDC hBmpDC;

			_ASSERT(nLcdDoubled == 1 || nLcdDoubled == 2 || nLcdDoubled == 4);
			hBmp    = CreateBitmap(131*nLcdDoubled,64*nLcdDoubled,1,1,NULL);
			hBmpDC  = CreateCompatibleDC(NULL);
			hOldBmp = (HBITMAP)SelectObject(hBmpDC,hBmp);
			BitBlt(hBmpDC, 0, 0, 131*nLcdDoubled, 64*nLcdDoubled, hLcdDC, 0, 0, SRCCOPY);
			SetClipboardData(CF_BITMAP,hBmp);
			SelectObject(hBmpDC,hOldBmp);
			DeleteDC(hBmpDC);
#endif
		}
		CloseClipboard();
	}
	return 0;
}

//
// ID_VIEW_RESET
//
static LRESULT OnViewReset()
{
	if (nState!=0) return 0;
	if (YesNoMessage(_T("Are you sure you want to press the Reset Button ?"))==IDYES)
	{
		SwitchToState(3);
		Chipset.pc = 0;
		Chipset.inte = TRUE;
		Chipset.Shutdn = FALSE;
		Chipset.SoftInt = TRUE;
		Reset();
		bInterrupt = TRUE;
		SwitchToState(0);
	}
	return 0;
}

//
// ID_VIEW_SETTINGS
//
static LRESULT OnViewSettings()
{
	ReadSettings();

	// in nState = 1 port2 file must be closed from document
	_ASSERT(nState != 1 || pbyPort2 == NULL);
//Para que los ports puedan configurarse
	SwitchToState(SM_INVALID);
	if (DialogBox(hApp, MAKEINTRESOURCE(IDD_SETTINGS), hWnd, (DLGPROC)SettingsProc) == -1)
		AbortMessage(_T("Settings Dialog Creation Error !"));
	RESTORECMDBAR
	WriteSettings();
//Para que los ports puedan configurarse
	SwitchToState(SM_RUN);
	return 0;
}

//
// ID_VIEW_SCRIPT
//
static LRESULT OnViewScript()
{
	CHAR cType = cCurrentRomType;
	if (nState!=0)
	{
		InfoMessage(_T("You cannot change the KML script when Emu48 is not running.\n\
Use the File,New menu item to create a new calculator."));
		return 0;
	}
	SwitchToState(1);
retry:
	if (!DisplayChooseKml(cType)) goto ok;
	if (!InitKML(szCurrentKml,FALSE))
	{
		if (AbortMessage(_T("This KML Script has errors.\nPress Ok to try to find a correct script."))==IDCANCEL)
		{
			if (szCurrentFilename[0])
				OnFileSave();
			else
				OnFileSaveAs();
			goto ok;
		}
		goto retry;
	}
ok:
	if (pbyRom) SwitchToState(0);
	return 0;
}

//
// ID_BACKUP_SAVE
//
static LRESULT OnBackupSave()
{
	UINT nOldState;
	if (pbyRom == NULL) return 0;
	nOldState = SwitchToState(1);
	SaveBackup();
	SwitchToState(nOldState);
	return 0;
}

//
// ID_BACKUP_RESTORE
//
static LRESULT OnBackupRestore()
{
	SwitchToState(1);
	RestoreBackup();
	if (pbyRom) SwitchToState(0);
	return 0;
}

//
// ID_BACKUP_DELETE
//
static LRESULT OnBackupDelete()
{
	ResetBackup();
	return 0;
}

//
// ID_OBJECT_LOAD
//
static LRESULT OnObjectLoad()
{
	static BOOL bWarning = TRUE;

	if (!Chipset.dispon)					// calculator off, turn on
	{
		// turn on HP
		KeyboardEvent(TRUE,0,0x8000);
		Sleep(200);
		KeyboardEvent(FALSE,0,0x8000);
	}

	if (nState!=0)
	{
		InfoMessage(_T("The emulator must be running to load an object."));
		return 0;
	}
	
	if (WaitForSleepState())				// wait for cpu SHUTDN then sleep state
	{
		InfoMessage(_T("The emulator is busy."));
		return 0;
	}

	while (nState!=nNextState) Sleep(0);
	_ASSERT(nState==3);
	
	if (bWarning)
	{
		UINT uReply = YesNoCancelMessage(
_T("Warning: Trying to load an object while the emulator is busy \n\
will certainly result in a memory lost. Before loading an object \n\
you should be sure that the HP48 is not doing anything. \n\
Do you want to see this warning next time you try to load an object ?"));
		switch (uReply)
		{
		case IDYES:
			break;
		case IDNO:
			bWarning = FALSE;
			break;
		case IDCANCEL:
			SwitchToState(0);
			return 0;
		}
	}
	
	if (!GetLoadObjectFilename())
	{
		SwitchToState(0);
		return 0;
	}
	
	if (!LoadObject(szBufferFilename))
	{
		SwitchToState(0);
		return 0;
	}

	SwitchToState(0);						// run state
	while (nState!=nNextState) Sleep(0);
	_ASSERT(nState==0);
	KeyboardEvent(TRUE,0,0x8000);
	Sleep(200);
	KeyboardEvent(FALSE,0,0x8000);
	while(Chipset.Shutdn == FALSE) Sleep(0);
	return 0;
}

//
// ID_OBJECT_SAVE
//
static LRESULT OnObjectSave()
{
	if (nState!=0)
	{
		InfoMessage(_T("The emulator must be running to save an object."));
		return 0;
	}
	
	if (WaitForSleepState())				// wait for cpu SHUTDN then sleep state
		return 0;

	while (nState!=nNextState) Sleep(0);
	_ASSERT(nState==3);

	if (!GetSaveObjectFilename())
	{
		SwitchToState(0);
		return 0;
	}
	
	SaveObject(szBufferFilename);
	
	SwitchToState(0);
	
	return 0;
}

//
// ID_TOOL_DISASM
//
#ifndef	UNDER_CE
static BOOL CALLBACK Disasm(HWND hDlg, UINT message, DWORD wParam, LONG lParam)
{
	static DWORD dwAddress, dwAddressMax;

	LONG  i;
	LPINT lpnCount;
	TCHAR  *cpStop,szAddress[256] = _T("0");
	
	switch (message)
	{
	case WM_INITDIALOG:
		// set fonts & cursor
		SendDlgItemMessage(hDlg,IDC_DISASM_MODULE,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_MAP,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_ROM,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_RAM,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_PORT1,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_PORT2,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
//		SendDlgItemMessage(hDlg,IDC_DISASM_MNEMONICS,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_HP,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_CLASS,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_MAP,BM_SETCHECK,1,0);
		SendDlgItemMessage(hDlg,IDC_DISASM_HP,BM_SETCHECK,1,0);
		SendDlgItemMessage(hDlg,IDC_ADDRESS,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_ADR,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_NEXT,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_COPY,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDCANCEL,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SetDlgItemText(hDlg,IDC_DISASM_ADR,szAddress);
		disassembler_map = MEM_MAP;			// disassemble with mapped modules
		disassembler_mode = HP_MNEMONICS;	// use HP mnemonics for disassembling
		dwAddress = strtoul(szAddress,&cpStop,16);
		dwAddressMax = 0x100000;			// greatest address (mapped mode)
		return TRUE;
	case WM_COMMAND:
		switch(wParam)
		{
		// decode radio buttons
		case IDC_DISASM_MAP:
			disassembler_map = MEM_MAP;
			dwAddressMax = 0x100000;
			return TRUE;
		case IDC_DISASM_ROM:
			disassembler_map = MEM_ROM;
			dwAddressMax = dwRomSize;
			return TRUE;
		case IDC_DISASM_RAM:
			disassembler_map = MEM_RAM;
			dwAddressMax = Chipset.Port0Size * 2048;
			return TRUE;
		case IDC_DISASM_PORT1:
			disassembler_map = MEM_PORT1;
			dwAddressMax = ((Chipset.cards_status & PORT1_PRESENT) != 0) ? (Chipset.Port1Size * 2048) : 0;
			return TRUE;
		case IDC_DISASM_PORT2:
			disassembler_map = MEM_PORT2;
			dwAddressMax = (cCurrentRomType=='X')
			             ? (Chipset.Port2Size * 2048)
						 : ((dwPort2Mask != 0) ? ((dwPort2Mask + 1) << 18) : (dwPort2Size * 2048));
			return TRUE;
		case IDC_DISASM_HP:
			disassembler_mode = HP_MNEMONICS;
			return TRUE;
		case IDC_DISASM_CLASS:
			disassembler_mode = CLASS_MNEMONICS;
			return TRUE;
		case IDOK:
			SendDlgItemMessage(hDlg,IDC_DISASM_ADR,EM_SETSEL,0,-1);
			GetDlgItemText(hDlg,IDC_DISASM_ADR,szAddress,sizeof(szAddress));
			// test if valid hex address
			for (i = 0; i < (LONG) strlen(szAddress); ++i)
			{
				if (isxdigit(szAddress[i]) == FALSE)
					return FALSE;
			}
			dwAddress = strtoul(szAddress,&cpStop,16);
			// no break
		case IDC_DISASM_NEXT:
			if (dwAddress >= dwAddressMax)
				return FALSE;
			i = wsprintf(szAddress,(dwAddress <= 99999) ? "%05lX   " : "%06lX  ",dwAddress);
			dwAddress = disassemble(dwAddress,&szAddress[i],VIEW_LONG);
			i = SendDlgItemMessage(hDlg,IDC_DISASM_WIN,LB_ADDSTRING,0,(LPARAM) szAddress);
			SendDlgItemMessage(hDlg,IDC_DISASM_WIN,LB_SELITEMRANGE,FALSE,MAKELPARAM(0,i));
			SendDlgItemMessage(hDlg,IDC_DISASM_WIN,LB_SETSEL,TRUE,i);
			SendDlgItemMessage(hDlg,IDC_DISASM_WIN,LB_SETTOPINDEX,i,0);
			return TRUE;
		case IDC_DISASM_COPY:
			// get number of selections
			if ((i = SendDlgItemMessage(hDlg,IDC_DISASM_WIN,LB_GETSELCOUNT,0,0)) == 0)
				break;						// no items selected

			if ((lpnCount = (LPINT) LocalAlloc(LMEM_FIXED,i * sizeof(INT))) != NULL)
			{
				HANDLE hClipObj;
				LONG j,lMem = 0;

				// get indexes of selected items
				i = SendDlgItemMessage(hDlg,IDC_DISASM_WIN,LB_GETSELITEMS,i,(LPARAM) lpnCount);
				for (j = 0;j < i;++j)		// scan all selected items
				{
					// calculate total amount of needed memory
					lMem += SendDlgItemMessage(hDlg,IDC_DISASM_WIN,LB_GETTEXTLEN,lpnCount[j],0) + 2;
				}
				// allocate clipboard data
				if ((hClipObj = LocalAlloc(GMEM_MOVEABLE,lMem + 1)) != NULL)
				{
					LPBYTE lpData;

					if (lpData = LocalLock(hClipObj))
					{
						for (j = 0;j < i;++j) // scan all selected items
						{
							lpData += SendDlgItemMessage(hDlg,IDC_DISASM_WIN,LB_GETTEXT,lpnCount[j],(LPARAM) lpData);
							*lpData++ = '\r';
							*lpData++ = '\n';
						}
						*lpData = 0;		// set end of string
						LocalUnlock(hClipObj); // unlock memory
					}

					if (OpenClipboard(hWnd))
					{
						if (EmptyClipboard())
							SetClipboardData(CF_TEXT,hClipObj);
						else
							LocalFree(hClipObj);
						CloseClipboard();
					}
					else					// clipboard open failed
					{
						LocalFree(hClipObj);
					}
				}
				LocalFree(lpnCount);		// free item table
			}
			return TRUE;
		case IDCANCEL:
			EndDialog(hDlg, wParam);
			return TRUE;
		}
		break;
	}
	return FALSE;
    UNREFERENCED_PARAMETER(lParam);
}
#endif
//
// ID_ABOUT
//
static BOOL CALLBACK About(HWND hDlg, UINT message, DWORD wParam, LONG lParam)
{
	
	switch (message)
	{
	case WM_INITDIALOG:
		SetDlgItemText(hDlg,IDC_VERSION,szNoTitle);
		SetDlgItemText(hDlg,IDC_LICENSE,szLicence);
		FULLSCREEN(hDlg);
		return TRUE;
	case WM_COMMAND:
		if ((LOWORD(wParam)==IDOK)||(LOWORD(wParam)==IDCANCEL))
		{
			EndDialog(hDlg, wParam);
			return TRUE;
		}
		break;
	}
	return FALSE;
    UNREFERENCED_PARAMETER(lParam);
}

static LRESULT OnToolDisasm()				// disasm dialogbox call
{
#ifndef	UNDER_CE
	if (pbyRom) SwitchToState(3);
	if (DialogBox(hApp, MAKEINTRESOURCE(IDD_DISASM), hWnd, (DLGPROC)Disasm) == -1)
		AbortMessage("Disassembler Dialog Box Creation Error !");
	if (pbyRom) SwitchToState(0);
#endif
	return 0;
}

static LRESULT OnAbout()
{
	if (DialogBox(hApp, MAKEINTRESOURCE(IDD_ABOUT), hWnd, (DLGPROC)About) == -1)
		AbortMessage(_T("About Dialog Box Creation Error !"));
	RESTORECMDBAR
	return 0;
}

static LRESULT OnLButtonDown(UINT nFlags, WORD x, WORD y)
{
	if (nState==0) MouseButtonDownAt(nFlags, x,y);
	return 0;
}

static LRESULT OnLButtonUp(UINT nFlags, WORD x, WORD y) 
{
	if (nState==0) MouseButtonUpAt(nFlags, x,y);
	return 0;
}

static LRESULT OnMouseMove(UINT nFlags, WORD x, WORD y)
{
	if (nState==0) MouseMovesTo(nFlags, x,y);
	return 0;
}

static LRESULT OnKeyDown(int nVirtKey, DWORD lKeyData)
{
	// call RunKey() only once (suppress autorepeat feature)
	if (nState==0 && (lKeyData & 0x40000000)==0)
		RunKey((BYTE)nVirtKey, TRUE);
	return 0;
}

static LRESULT OnKeyUp(int nVirtKey, DWORD lKeyData)
{
	if (nState==0) RunKey((BYTE)nVirtKey, FALSE);
	return 0;
	UNREFERENCED_PARAMETER(lKeyData);
}

static LRESULT CALLBACK MainWndProc(HWND hWindow, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_CREATE:          return OnCreate(hWindow);
	case WM_DESTROY:         return OnDestroy(hWindow);
	case WM_PAINT:           return OnPaint(hWindow);
#ifdef	UNDER_CE
	case WM_HELP:
		{
			SHELLEXECUTEINFO	info;

			memset(&info,0,sizeof(SHELLEXECUTEINFO));
			info.cbSize=sizeof(SHELLEXECUTEINFO);
			info.hwnd=hWindow;
			info.lpFile=_T("peghelp.exe");
			info.lpParameters=_T("emu48.htm");
			info.lpDirectory=szEmu48Directory;
			info.hInstApp=hApp;
			ShellExecuteEx(&info);
		}
		break;
#endif
#ifndef	UNDER_CE
	case WM_DROPFILES:       return OnDropFiles((HANDLE)wParam);
#endif
	case WM_ENTERMENULOOP:
		SwitchToState(SM_SLEEP);
		break;
	case WM_EXITMENULOOP:
		SwitchToState(SM_RUN);
		break;
	case WM_ACTIVATE:
		if (LOWORD(wParam)==WA_INACTIVE) 
		{
			break;
		}
	case WM_QUERYNEWPALETTE:
		if (hPalette)
		{
			SelectPalette(hWindowDC, hPalette, FALSE);
			if (RealizePalette(hWindowDC))
			{
				InvalidateRect(hWindow,NULL,TRUE);
				return TRUE;
			}
		}
		return FALSE;
	case WM_PALETTECHANGED:
		if ((HWND)wParam == hWindow) break;
		if (hPalette)
		{
			SelectPalette(hWindowDC, hPalette, FALSE);
			if (RealizePalette(hWindowDC))
			{
				// UpdateColors(hWindowDC);
				InvalidateRect (hWnd, (LPRECT) (NULL), 1);
			}
		}
		return FALSE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_SWAP:		return SwapDocument();
		case ID_FILE_NEW:      return OnFileNew();
		case ID_FILE_OPEN:     return OnFileOpen();
		case ID_FILE_SAVE:     return OnFileSave();
		case ID_FILE_SAVEAS:   return OnFileSaveAs();
		case ID_FILE_CLOSE:    return OnFileClose();
		case ID_FILE_EXIT:     return OnFileExit();
		case ID_STACK_COPY:    return OnStackCopy();
		case ID_STACK_PASTE:   return OnStackPaste();
		case ID_VIEW_COPY:     return OnViewCopy();
		case ID_VIEW_RESET:    return OnViewReset();
		case ID_VIEW_SETTINGS: return OnViewSettings();
		case ID_VIEW_SCRIPT:   return OnViewScript();
		case ID_BACKUP_SAVE:   return OnBackupSave();
		case ID_BACKUP_RESTORE:return OnBackupRestore();
		case ID_BACKUP_DELETE: return OnBackupDelete();
		case ID_OBJECT_LOAD:   return OnObjectLoad();
		case ID_OBJECT_SAVE:   return OnObjectSave();
		case ID_TOOL_DISASM:   return OnToolDisasm();
#ifndef	UNDER_CE
		case ID_TOOL_DEBUG:    return OnToolDebug();
#endif
		case ID_ABOUT:         return OnAbout();
		}
		break;
	case WM_SYSCOMMAND:
		switch (LOWORD(wParam))
		{
		case SC_CLOSE: return OnFileExit();
		}
		break;
	case WM_RBUTTONDOWN:
	case WM_LBUTTONDOWN: return OnLButtonDown(wParam, LOWORD(lParam), HIWORD(lParam));
	case WM_LBUTTONUP:   return OnLButtonUp(wParam, LOWORD(lParam), HIWORD(lParam));
	case WM_MOUSEMOVE:   return OnMouseMove(wParam, LOWORD(lParam), HIWORD(lParam));
	case WM_KEYUP:       return OnKeyUp((int)wParam, lParam);
	case WM_KEYDOWN:     return OnKeyDown((int)wParam, lParam);
	}
	return DefWindowProc(hWindow, uMsg, wParam, lParam);
}

static BOOL FlushMessages(LPMSG msg)
{
	while (PeekMessage(msg, NULL, 0, 0, PM_REMOVE))
	{
		if (msg->message == WM_QUIT) return TRUE;
		TranslateMessage(msg);
		DispatchMessage(msg);
	}
	return FALSE;
}

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPTSTR lpCmdLine, int nCmdShow)
{
	MSG msg;
	WNDCLASS wc;
	RECT rectWindow;
	DWORD style;
#ifndef	UNDER_CE
	HSZ hszService, hszTopic;				// variables for DDE server
#endif

	hWnd=FindWindow(_T("CEmu48"),NULL);
	if (hWnd)
	{
		SetForegroundWindow (hWnd);    
		return 0;
	}

	hApp = hInst;

	memset(&wc,0,sizeof(WNDCLASS));
#ifndef	UNDER_CE
	wc.style = CS_BYTEALIGNCLIENT;
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
#endif
	StartCmdLine(lpCmdLine);
	wc.lpfnWndProc = (WNDPROC)MainWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_EMU48));
	wc.hbrBackground = (struct HBRUSH__ *)(COLOR_APPWORKSPACE+1);
	wc.lpszClassName = _T("CEmu48");
	
	if (!RegisterClass(&wc))
	{
		AbortMessage(
_T("CEmu48 class registration failed.\n\
This application will now terminate."));
		return FALSE;
	}
	// Create window
	
#if defined(UNDER_CE)
	WinCEStartup();
#ifdef UNDER_CE_PPC
	rectWindow.left   = 0;
	rectWindow.top    = 0;
	rectWindow.right  = 240;
	rectWindow.bottom = 320-26;
#else
	rectWindow.bottom = 320-16;
#endif
#else
	rectWindow.left   = 0;
	rectWindow.top    = 0;
	rectWindow.right  = 256;
	rectWindow.bottom = 0;
#endif
#ifndef	UNDER_CE
	AdjustWindowRect(&rectWindow, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);
	hWnd = CreateWindow(_T("CEmu48"), _T("Emu48"),
		WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED,
		CW_USEDEFAULT, CW_USEDEFAULT,
		rectWindow.right  - rectWindow.left,
		rectWindow.bottom - rectWindow.top,
		NULL,NULL,hApp,NULL
		);
#else
#ifdef	UNDER_CE_PPC
	style=WS_VISIBLE;
#else
	style=WS_OVERLAPPED | WS_SYSMENU;
#endif
	hWnd = CreateWindow(_T("CEmu48"), _T("HP 48/49G Emulator"),
		style,
		0,0,
		rectWindow.right  - rectWindow.left,
		rectWindow.bottom - rectWindow.top,
		NULL,NULL,hApp,NULL
		);
#endif
	if (hWnd == NULL)
	{
		AbortMessage(_T("Window creation failed.\n"));
		return FALSE;
	}
	
//	ShowWindow(hWnd, nCmdShow);

//	InfoMessage(_T("1st Flush Message"));
	
	if (FlushMessages(&msg)) return msg.wParam;

//	InfoMessage(_T("After 1st Flush Message"));
	
	// initialization
	QueryPerformanceFrequency(&lFreq);		// init high resolution counter
	QueryPerformanceFrequency(&lAppStart);

//Chapuza temporar
	GetCurrentDirectory(sizeof(szCurrentDirectory), szCurrentDirectory);
//	InfoMessage(_T("After GetCurrentDir"));
	
	ReadSettings();
	
//	InfoMessage(_T("After ReadSettings"));

	UpdateWindowStatus();
//	InfoMessage(_T("After UpdateWindowStatus"));

	// create auto event handle
	hEventShutdn = CreateEvent(NULL,FALSE,FALSE,NULL);
	if (hEventShutdn == NULL)
	{
		AbortMessage(_T("Event creation failed."));
		return FALSE;
	}

	nState     = SM_RUN;    // init state must be <> nNextState
	nNextState = SM_INVALID;
	hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)&WorkerThread, NULL, 0, &lThreadId);
	if (hThread == NULL)
	{
		CloseHandle(hEventShutdn);			// close event handle
		AbortMessage(_T("Thread creation failed."));
		return FALSE;
	}
	// initialize DDE server
#ifndef	UNDER_CE
	if (DdeInitialize(&idDdeInst,(PFNCALLBACK) &DdeCallback,
                      APPCLASS_STANDARD |
                      CBF_FAIL_EXECUTES | CBF_FAIL_ADVISES |
                      CBF_SKIP_REGISTRATIONS | CBF_SKIP_UNREGISTRATIONS,0))
	{
		AbortMessage("Could not initialize server!");
		return FALSE;
	}
#endif
	if (ArgC() >= 2)						// use decoded parameter line
	{
		TCHAR szTemp[256] = _T("Loading ");
		lstrcat(szTemp, ArgV(1));			// use decoded parameter line
		SetWindowTitle(szTemp);
		if (FlushMessages(&msg)) return msg.wParam;
		if (OpenDocument(ArgV(1)))		// use decoded parameter line
			goto start;
	}

	GetPrivateProfileString(_T("Files"),_T("LastDocument"),_T(""),szBufferFilename,sizeof(szBufferFilename),EMU48_INI);
	GetPrivateProfileString(_T("Files"),_T("OtherDocument"),_T(""),szOtherFilename,sizeof(szOtherFilename),EMU48_INI);

	if (szBufferFilename[0])
	{
		TCHAR szTemp[256] = _T("Loading ");
		_tcscat(szTemp, szBufferFilename);
		SetWindowTitle(szTemp);
//		MessageBox(hWnd, szTemp, _T("Loading"), MB_OK);
		if (FlushMessages(&msg)) return msg.wParam;
		if (OpenDocument(szBufferFilename))
			goto start;
	}
	
	SetWindowTitle(_T("New Document"));
	if (FlushMessages(&msg)) return msg.wParam;
	if (NewDocument())
	{
		SetWindowTitle(_T("Untitled"));
		goto start;
	}
	
	ResetDocument();
	
start:
	InitializeCriticalSection(&csGDILock);
	InitializeCriticalSection(&csKeyLock);
	InitializeCriticalSection(&csIOLock);
	InitializeCriticalSection(&csT1Lock);
	InitializeCriticalSection(&csT2Lock);

	// init clipboard format and name service
#ifndef	UNDER_CE
	uCF_HpObj = RegisterClipboardFormat(CF_HPOBJ);
	hszService = DdeCreateStringHandle(idDdeInst,szAppName,0);
	hszTopic   = DdeCreateStringHandle(idDdeInst,szTopic,0);
	DdeNameService(idDdeInst,hszService,NULL,DNS_REGISTER);
#endif

	if (pbyRom) 
		SwitchToState(SM_RUN);
	UpdateWindowStatus();
	while (GetMessage(&msg, NULL, 0, 0))
	{
#ifndef	UNDER_CE
		if(hDlgDebug == NULL || !IsDialogMessage(hDlgDebug,&msg))
#endif
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

#ifndef	UNDER_CE
	// clean up DDE server
	DdeNameService(idDdeInst, hszService, NULL, DNS_UNREGISTER);
	DdeFreeStringHandle(idDdeInst, hszService);
	DdeFreeStringHandle(idDdeInst, hszTopic);
	DdeUninitialize(idDdeInst);
#else
	WinCETerminate();
#endif

	DeleteCriticalSection(&csGDILock);
	DeleteCriticalSection(&csKeyLock);
	DeleteCriticalSection(&csIOLock);
	DeleteCriticalSection(&csT1Lock);
	DeleteCriticalSection(&csT2Lock);
	
	_ASSERT(nState == 2);					// set by WM_DESTROY
	CloseHandle(hEventShutdn);				// close event handle
	ResetDocument();
	ResetBackup();
	_ASSERT(pbyPort2 == NULL);				// port2 file unmapped
	_ASSERT(pKml == NULL);					// KML script not closed
	_ASSERT(szTitle == NULL);				// freed allocated memory
	_ASSERT(hPalette == NULL);				// freed resource memory
	
	return msg.wParam;
	UNREFERENCED_PARAMETER(lpCmdLine);
	UNREFERENCED_PARAMETER(hPrevInst);
}

//################
//#
//#    Clipboard Tool
//#
//################

VOID CopyItemsToClipboard(HWND hWnd)		// save selected Listbox Items to Clipboard
{
#ifndef	UNDER_CE
	LONG  i;
	LPINT lpnCount;
	
	// get number of selections
	if ((i = SendMessage(hWnd,LB_GETSELCOUNT,0,0)) == 0)
		return;								// no items selected

	if ((lpnCount = (LPINT) LocalAlloc(LMEM_FIXED,i * sizeof(INT))) != NULL)
	{
		HANDLE hClipObj;
		LONG j,lMem = 0;

		// get indexes of selected items
		i = SendMessage(hWnd,LB_GETSELITEMS,i,(LPARAM) lpnCount);
		for (j = 0;j < i;++j)				// scan all selected items
		{
			// calculate total amount of needed memory
			lMem += SendMessage(hWnd,LB_GETTEXTLEN,lpnCount[j],0) + 2;
		}
		// allocate clipboard data
		if ((hClipObj = GlobalAlloc(GMEM_MOVEABLE,lMem + 1)) != NULL)
		{
			LPBYTE lpData;

			if (lpData = GlobalLock(hClipObj))
			{
				for (j = 0;j < i;++j)		// scan all selected items
				{
					lpData += SendMessage(hWnd,LB_GETTEXT,lpnCount[j],(LPARAM) lpData);
					*lpData++ = '\r';
					*lpData++ = '\n';
				}
				*lpData = 0;				// set end of string
				GlobalUnlock(hClipObj);		// unlock memory
			}

			if (OpenClipboard(hWnd))
			{
				if (EmptyClipboard())
					SetClipboardData(CF_TEXT,hClipObj);
				else
					GlobalFree(hClipObj);
				CloseClipboard();
			}
			else							// clipboard open failed
			{
				GlobalFree(hClipObj);
			}
		}
		LocalFree(lpnCount);				// free item table
	}
	return;
#endif
}

