#include <windows.h>
#include <windowsx.h>
#include "registry.h"
#include <string.h>
#include <winreg.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

BOOL iWriteRegistry(HKEY , LPCTSTR lpSubKey, LPCTSTR lpValueName, 
							   DWORD dwType, LPBYTE lpData, DWORD cbData);
BOOL iReadRegistry(HKEY hKey, LPCTSTR lpSubKey, LPCTSTR lpValueName,
							  LPDWORD lpType, BYTE **lppData, LPDWORD lpcbData);

BOOL writeRegistryString(HKEY hKey, LPCTSTR lpSubKey, LPCTSTR lpValueName, LPCTSTR lpData)
{
	// When we pass the data size, make sure to include the terminating NULL
	return iWriteRegistry(hKey, lpSubKey, lpValueName, REG_SZ, (LPBYTE)lpData, (lstrlen(lpData) + 1) * sizeof(TCHAR));
}

BOOL writeRegistryInt(HKEY hKey, LPCTSTR lpSubKey, LPCTSTR lpValueName, DWORD dwValue)
{
	return iWriteRegistry(hKey, lpSubKey, lpValueName, REG_DWORD, (LPBYTE)&dwValue, sizeof(DWORD));
}

BOOL readRegistryString(HKEY hKey, LPCTSTR lpSubKey, LPCTSTR lpValueName, 
							 LPTSTR pstrValue, LPCTSTR lpDefault)
{
	BOOL bRes;
	LPBYTE lpData;
	DWORD cbData;
	DWORD dwType;

	cbData = 0;
	lpData = 0;

	bRes = iReadRegistry(hKey, lpSubKey, lpValueName, &dwType, &lpData, &cbData);
//	MessageBox(0, lpValueName, _T("just ReadRegistry"), MB_OK);
	if (bRes)
	{
//		MessageBox(0, lpValueName, _T("found in Registry"), MB_OK);
		if (cbData)
		{
			lstrcpy(pstrValue, (LPTSTR)lpData);
		}
		else if (lpDefault)
		{
//			MessageBox(0, lpValueName, _T("Returning as default"), MB_OK);
//			pstrValue = (LPTSTR)malloc(lstrlen((LPCTSTR)lpDefault));
//			if (IsBadWritePtr(pstrValue, lstrlen((LPCTSTR)lpDefault)))
//				return FALSE;
			lstrcpy(pstrValue, (LPCTSTR)lpDefault);
		}
	}
	else
	{
		if (lpDefault)
		{
//			pstrValue = (LPTSTR)malloc((lstrlen(lpDefault) + 1) + sizeof(TCHAR));
//			if (IsBadWritePtr(pstrValue, (lstrlen(lpDefault) + 1) + sizeof(TCHAR)))
//				return FALSE;
			lstrcpy(pstrValue, (LPTSTR)lpDefault);
		}
	}
	return bRes;
}

BOOL readRegistryInt(HKEY hKey, LPCTSTR lpSubKey, LPCTSTR lpValueName, LPDWORD lpdwValue,
		DWORD dwDefault)
{
	DWORD dwType;
	DWORD cbData;
	LPBYTE lpData;
	BOOL bRes;

	lpData = 0;
	cbData = 0;

	bRes =  iReadRegistry(hKey, lpSubKey, lpValueName, &dwType, &lpData, &cbData);
	if (bRes)
	{
		if (cbData)
		{
			if (dwType == REG_DWORD)
				*lpdwValue = (DWORD)(*((DWORD *)lpData));
			else
				bRes = FALSE;
			free(lpData);
		}
		else 
			*lpdwValue = dwDefault;
	}
	return bRes;
}

BOOL iWriteRegistry(HKEY hKey, LPCTSTR lpSubKey, LPCTSTR lpValueName, 
							   DWORD dwType, LPBYTE lpData, DWORD cbData)
{
	HKEY hLocalKey;
	DWORD dwDist;
	long lRes;

	// Open the key, or create it if it doesn't exist
	m_lLastRegError = RegCreateKeyEx(hKey, lpSubKey, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE,
			NULL, &hLocalKey, &dwDist);

	if ((m_lLastRegError == ERROR_SUCCESS) || (m_lLastRegError == ERROR_FILE_NOT_FOUND))
	{
		// Set the value, and close the registry key
		m_lLastRegError = RegSetValueEx(hLocalKey, lpValueName, 0, dwType, (LPBYTE)((void *)lpData), cbData);
		lRes = RegCloseKey(hLocalKey);
		if ((m_lLastRegError == ERROR_SUCCESS) || (m_lLastRegError == ERROR_FILE_NOT_FOUND))
			m_lLastRegError = lRes;
	}

	if ((m_lLastRegError == ERROR_SUCCESS) || (m_lLastRegError == ERROR_FILE_NOT_FOUND))
		return TRUE;

	return FALSE;
}

BOOL iReadRegistry(HKEY hKey, LPCTSTR lpSubKey, LPCTSTR lpValueName,
							  LPDWORD lpType, BYTE **lppData, LPDWORD lpcbData)
{
	HKEY hLocalKey;
	long lRes;
	DWORD dwDist;

	// Open the key, or create it if it doesn't exist
	m_lLastRegError = RegCreateKeyEx(hKey, lpSubKey, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_READ,
			NULL, &hLocalKey, &dwDist);
	if ((m_lLastRegError == ERROR_SUCCESS) || (m_lLastRegError == ERROR_FILE_NOT_FOUND))
	{
		// If lpData is null, query the registry to find out how much memory to allocate for the data...
		if (!(*lppData))
		{
			// Get the memory required
			m_lLastRegError = RegQueryValueEx(hLocalKey, lpValueName, NULL, lpType, NULL, lpcbData);
			if ((m_lLastRegError == ERROR_SUCCESS) || 
					(m_lLastRegError == ERROR_FILE_NOT_FOUND))
			{
				if (lpcbData)
				{
					// Allocate the memory for the data
					*lppData = (LPBYTE)malloc(*lpcbData);
					if (IsBadWritePtr(*lppData, *lpcbData))
					{
						lRes = RegCloseKey(hLocalKey);
						return FALSE;
					}
				}
				else
				{
					// Couldn't find the registry entry
					m_lLastRegError = RegCloseKey(hLocalKey);
					return TRUE;
				}
			}
			else
			{
				// Windows CE can return ERROR_INVALID_PARAMETER
				// if the registry key doesn't exist.
				if (m_lLastRegError == ERROR_INVALID_PARAMETER)
				{
					*lpcbData = 0;
					lRes = RegCloseKey(hLocalKey);
					return TRUE;
				}
				else
				{
					// Panic!
					lRes = RegCloseKey(hLocalKey);
					return FALSE;
				}
			}
		}
		// Query the value from the registry
		m_lLastRegError = RegQueryValueEx(hLocalKey, lpValueName, NULL, lpType, *lppData, lpcbData);
	}

	if ((m_lLastRegError == ERROR_SUCCESS) || (m_lLastRegError == ERROR_FILE_NOT_FOUND))
		return TRUE;

	return FALSE;
}